/*
 * Copyright (c) 2007, 2008  Paul Vojta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 */

/* $XFree86: xc/programs/Xserver/hw/xfree86/input/synaptics/synaptics.c,v 1.0 2003/11/15 01:01:01 vojta Exp $ */

/* To do:
	options processing (xf86AddModuleInfo mouse.c)
	list of externals (xf86LoaderReqSymLists microtouch.c)
	Proximity events?  (microtouch.c)
 */

/*========================================================================*\
 *
 *	Header stuff (include files, definitions, function declarations)
 *
\*========================================================================*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <xf86.h>
#include <xf86Xinput.h>
#include <xf86_OSproc.h>
#include <xisb.h>
#include <mipointer.h>
#include <xf86Module.h>
#include <extnsionst.h>
#include <synpt.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>

#if HAVE_LINUX_INPUT_H
# include <linux/input.h>
#endif

#include <errno.h>

#define	Number(x)	(sizeof (x) / sizeof *(x))

#define	SYN_N_BUTTONS	7		/* number of mouse buttons */

static	InputInfoPtr	pre_init(InputDriverPtr, IDevPtr, int);
static	Bool		tp_control(DeviceIntPtr, int);
static	void		tp_read(InputInfoPtr);
static	Bool		tp_convert(InputInfoPtr, int, int, int, int, int, int,
			  int, int, int *, int *);
static	pointer		module_setup(pointer, pointer, int *, int *);
static	void		module_teardown(pointer);

static	DISPATCH_PROC(extn_main);
static	DISPATCH_PROC(extn_swapped_main);
static	void		extn_closedown(ExtensionEntry *);

/*========================================================================*\
 *
 *	Data structures
 *
\*========================================================================*/

struct optl {
	const char	*name;
	int		type;
	int		value;
	int		deflt;
	int		min, max;
};

#define	OPT_FLAG	0
#define	OPT_BYTE	1
#define	OPT_SHORT	2

#ifdef offsetof
# define o(field)	offsetof(SynptControlRec, field)
#else
# define o(field)	((size_t) &((SynptControlRec *)0)->field)
#endif

	/* The minimum and maximum values below occur also in opt_valid() */

const struct optl	optlist[]	= {
	{"Enabled",	OPT_FLAG,	SYN_ENABLED,	TRUE,	0,	0},
	{"HighRate",	OPT_FLAG,	SYN_FAST,	TRUE,	0,	0},
	{"Emulate3Buttons", OPT_FLAG,	SYN_EMU3,	FALSE,	0,	0},
	{"Emulate3Timeout", OPT_SHORT,	o(time_emu3),	55,	25,	5000},
	{"TapToClick",	OPT_FLAG,	SYN_TAPCLICK,	FALSE,	0,	0},
	{"TapAndDrag",	OPT_FLAG,	SYN_TAPDRAG,	FALSE,	0,	0},
	{"EdgeMotionUseAlways", OPT_FLAG, SYN_EDGEMOALWYS, FALSE, 0,	0},
	{"LeftEdgeScroll", OPT_FLAG,	SYN_L_EDGE_SCRL, FALSE,	0,	0},
	{"RightEdgeScroll", OPT_FLAG,	SYN_R_EDGE_SCRL, FALSE,	0,	0},
	{"TopEdgeScroll", OPT_FLAG,	SYN_T_EDGE_SCRL, FALSE,	0,	0},
	{"BottomEdgeScroll", OPT_FLAG,	SYN_B_EDGE_SCRL, FALSE,	0,	0},
	{"HorizTwoFingerScroll", OPT_FLAG, SYN_H_2FNG_SCRL, FALSE, 0,	0},
	{"VertTwoFingerScroll", OPT_FLAG, SYN_V_2FNG_SCRL, FALSE, 0,	0},
	{"MaxTapTime",	OPT_SHORT,	o(time_maxtap),	180,	25,	5000},
	{"TapDuration",	OPT_SHORT,	o(time_tapdur),	155,	25,	950},
	{"HorizScrollDelta", OPT_SHORT,	o(h_scroll_delta), 20,	4,	500},
	{"VertScrollDelta", OPT_SHORT,	o(v_scroll_delta), 20,	4,	500},
	{"FingerHigh",	OPT_BYTE,	o(finger_high),	29,	1,	255},
	{"FingerLow",	OPT_BYTE,	o(finger_low),	20,	1,	255},
	{"EdgeMotionMinZ", OPT_BYTE,	o(edge_min_z),	50,	0,	255},
	{"EdgeMotionMaxZ", OPT_BYTE,	o(edge_max_z),	100,	0,	255},
	{"EdgeMotionMinSpeed", OPT_BYTE, o(edge_min_speed), 24,	0,	255},
	{"EdgeMotionMaxSpeed", OPT_BYTE, o(edge_max_speed), 90,	0,	255},
	{"TapButton1",	OPT_BYTE,	o(tap_button[0]), 1, 0, SYN_N_BUTTONS},
	{"TapButton2",	OPT_BYTE,	o(tap_button[1]), 2, 0, SYN_N_BUTTONS},
	{"TapButton3",	OPT_BYTE,	o(tap_button[2]), 3, 0, SYN_N_BUTTONS},
	{"ULCornerButton", OPT_BYTE,	o(ul_button),	0,   0, SYN_N_BUTTONS},
	{"URCornerButton", OPT_BYTE,	o(ur_button),	0,   0, SYN_N_BUTTONS},
	{"LLCornerButton", OPT_BYTE,	o(ll_button),	0,   0, SYN_N_BUTTONS},
	{"LRCornerButton", OPT_BYTE,	o(lr_button),	0,   0, SYN_N_BUTTONS},
};

#undef o

/* Default values for locations of various edges */
static const SynptGeomRec default_motion_geometry = {10, 10, 13, 13};
static const SynptGeomRec default_scroll_geometry = {15, 15, 21, 21};
static const SynptGeomRec default_corner_geometry = {21, 21, 28, 26};

struct int_geom {
	int		l, r, t, b;
};

typedef struct {
	pointer		buffer;		/* for Xisb */
	InputInfoPtr	infp;
	int		protocol;
	int		protobak;	/* always stores the *real* protocol */

	union {
	  unsigned char	packet[6];	/* for psaux-mode packets */
#if HAVE_LINUX_INPUT_H
	  struct input_event ev;	/* for Linux input events */
#endif
	} u;
	int		packet_size;	/* how many bytes we have so far */

	/* for all packets */
	unsigned int	x, y, z, w;
	unsigned int	ev_buttons, prev_ev_buttons;
	unsigned char	ev_fingers;

	ExtensionEntry	*extn;
	SynptControlRec	opt;		/* most options go here */
	unsigned int	prev_x;
	unsigned int	prev_y;
	unsigned char	pressing;
	unsigned char	fingers;	/* current number of fingers touching */
	unsigned char	max_fingers;
	unsigned char	dx_low_bits;
	unsigned char	dy_low_bits;
	unsigned char	wheel_bits;
	unsigned char	in_sync;
	double		speed;
	double		speed_smooth;	/* 1/3 or 1/6 (smooth out speed) */
	double		dist2speed;	/* .02 or .04 (controls acceleration) */
	int		rate_factor;	/* 2 if slow, 1 if fast packet rate */
	int		emu3state;	/* state information for emu3buttons */
	OsTimerPtr	emu3timer;
	OsTimerPtr	gesture_timer;	/* packets before gesture timeout */
	struct int_geom	mo_ig;		/* location of edge (motion) */
	struct int_geom	scrl_ig;	/* location of edge (scrolling) */
	struct int_geom	cor_ig;		/* location of edge (corners) */
	unsigned int	gx, gy;		/* saved location of tap */
	unsigned char	scroll_type;	/* flags SCROLL_* */
	unsigned char	gbits;		/* button bits if gesturing */
	unsigned char	hard_buttons;	/* non-gesture mouse buttons */
	unsigned char	buttons;	/* mouse buttons (actual and gestured)*/
	unsigned char	hw_version;	/* hardware major version number */
	unsigned char	hw_subvers;	/* hardware minor version number */
	unsigned short	cap_bits;	/* capability bits (found by probing) */
} SynapticsPrivateRec, *SynapticsPrivatePtr;

/* Protocol symbols are defined in synpt.h, except for ... */
#define	SYN_PROTO_OFF	4	/* Used to implement ~SYN_ENABLED flag */

/* Bits for synp->ev_buttons */
#define	SYN_BTN_1	(1<<0)
#define	SYN_BTN_2	(1<<1)
#define	SYN_BTN_U	(1<<2)
#define	SYN_BTN_D	(1<<3)

/* Bits for synp->scroll_type */
#define	SCROLL_V_LEFT	(1<<0)
#define	SCROLL_V_RIGHT	(1<<1)
#define	SCROLL_H_TOP	(1<<2)
#define	SCROLL_H_BOTTOM	(1<<3)
#define	SCROLL_V_2FNGRS	(1<<4)
#define	SCROLL_H_2FNGRS	(1<<5)

#define	SCROLL_V_ALL	(SCROLL_V_LEFT | SCROLL_V_RIGHT | SCROLL_V_2FNGRS)
#define	SCROLL_H_ALL	(SCROLL_H_TOP | SCROLL_H_BOTTOM | SCROLL_H_2FNGRS)

#define	SCROLL_V_EDGE	(SCROLL_V_LEFT | SCROLL_V_RIGHT)
#define	SCROLL_H_EDGE	(SCROLL_H_TOP | SCROLL_H_BOTTOM)

static InputDriverRec synaptics = {	/* defined in xf86Xinput.h */
	/* driverVersion */		0,
	/* driverName */		"synaptics",
	/* Identify */			NULL,
	/* PreInit */			pre_init,
	/* UnInit */			NULL,
	/* module */			NULL,
	/* refCount */			0,
};

static XF86ModuleVersionInfo VersionRec =
{
    "synaptics",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
    XORG_VERSION_CURRENT,
    1, 0, 0,
    ABI_CLASS_XINPUT,
    ABI_XINPUT_VERSION,
    MOD_CLASS_XINPUT,
    {0, 0, 0, 0}	/* signature, to be patched into the file by a tool */
};

XF86ModuleData synapticsModuleData = {
	/* vers */			&VersionRec,
	/* setup */			module_setup,
	/* teardown */			module_teardown,
};

/*
 *	From Synaptics manual
 */

#define	SYN_XMIN 1472
#define	SYN_XMAX 5472
#define	SYN_YMIN 1408
#define	SYN_YMAX 4448

#define	SYN_XSIZE	(SYN_XMAX - SYN_XMIN)
#define	SYN_YSIZE	(SYN_YMAX - SYN_YMIN)

/*========================================================================*\
 *
 *	Procedures specific to raw protocol
 *
\*========================================================================*/

static Bool
ps2_send(InputInfoPtr infp, const unsigned char *seq, size_t n,
	const char *giveup)
{
	const unsigned char	*p;
	int			count;

	for (p = seq; p < seq + n; ++p) {
	    for (count = 0;; ++count) {	/* loop over retries */
		unsigned char c;

		if (count >= 5)	/* max retries */
		    return FALSE;

		xf86WriteSerial(infp->fd, p, 1);
		usleep(10000);
		if (xf86WaitForInput(infp->fd, 200000) <= 0) {
		    xf86Msg(X_PROBED, "Synaptics touchpad refused byte 0x%x\n",
		      *p);
		    return FALSE;
		}
		xf86ReadSerial(infp->fd, &c, 1);

		if (c == 0xfa) /* ack */
		    break;

		if (c == 0xfe) /* resend */
		    continue;

		if (c == 0xfc) {	/* error */
		    xf86Msg(X_PROBED, giveup);
		    return FALSE;
		}
	    }
	}

	return TRUE;
}

static Bool
send_mode_enable(InputInfoPtr infp, const char *msg)
{
	SynapticsPrivatePtr synp	= infp->private;
	static const unsigned char seq0[] = {	232, 2,	  /* 10 */
						232, 0,	  /* 00 */
						232, 0,	  /* 00 */
						232, 0,	  /* 00 */
						243, 20,  /* set mode byte */
						244, };	  /* enable */
	unsigned char seq[sizeof seq0];

	memcpy(seq, seq0, sizeof seq);
	if (synp->cap_bits) seq[7] |= 1;	/* set wmode bit */
	if (synp->opt.flags & SYN_FAST) seq[1] |= 1;  /* fast data rate */

	return ps2_send(infp, seq, sizeof seq, msg);
}

static Bool
init_synaptics(InputInfoPtr infp)
{
	SynapticsPrivatePtr synp	= infp->private;

	static const unsigned char s1[]	= { 246, };	/* set defaults */
							/* and disable */
	static const unsigned char s2[]	= { 232, 0,	/* 00 */
					    232, 0,	/* 00 */
					    232, 0,	/* 00 */
					    232, 0,	/* 00 */
					    233, };	/* status request */
	static const unsigned char s3[]	= { 232, 0,	/* 00 */
					    232, 0,	/* 00 */
					    232, 0,	/* 00 */
					    232, 2,	/* 10 */
					    233, };	/* get capability bits*/
	unsigned char answer[sizeof s2 + 3];
	int len;

	/* set defaults and disable */
	xf86WriteSerial(infp->fd, s1, sizeof s1);
	usleep(30000);
	xf86FlushInput(infp->fd);
	xf86ReadSerial(infp->fd, answer, sizeof answer);

	/* query for Synaptics type */
	xf86WriteSerial(infp->fd, s2, sizeof s2);
	usleep(30000);
	len = xf86ReadSerial(infp->fd, answer, sizeof answer);
	if (len != sizeof s2 + 3 || answer[sizeof s2 + 1] != 0x47) {
	    xf86Msg(X_PROBED, "Synaptics touchpad not in use; failing module startup\n");
	    xf86Msg(X_PROBED, "Reason for the above:  %d, %02x\n",
	      len, answer[sizeof s2 + 1]);
	    return FALSE;
	}
	xf86Msg(X_PROBED, "Synaptics touchpad version %d.%d detected\n",
	  synp->hw_version = answer[sizeof s2 + 2] & 0x0f,
	  synp->hw_subvers = answer[sizeof s2]);

	if ((answer[sizeof s2 + 2] & 0x0f) >= 4) {
	    xf86WriteSerial(infp->fd, s3, sizeof s3);	/* cap. bits */
	    usleep(30000);
	    len = xf86ReadSerial(infp->fd, answer, sizeof answer);
	    if (len == sizeof s3 + 3 && answer[sizeof s3 + 1] == 0x47
	      && (answer[sizeof s3] & 0x80) != 0) {
		synp->cap_bits = (unsigned int) answer[sizeof s3] << 8
		  | answer[sizeof s3 + 2];
		xf86Msg(X_PROBED, "Synaptics extended capability bits = %04x\n",
		  synp->cap_bits);
	    }
	}

	return send_mode_enable(infp,
	  "Synaptics touchpad flagged error on init; giving up\n");
}


/*========================================================================*\
 *
 *	Procedures for probing for event protocol
 *
\*========================================================================*/

/*
 *	Sorting routine (this is probably overkill).
 */

static void
quicksort(unsigned int *arr, size_t n)
{
	unsigned int pivot;
	size_t pivindex;
	unsigned int *arrp;

	if (n <= 1)
	    return;

	pivot = arr[n / 2];
	arr[n / 2] = arr[n - 1];
	pivindex = 0;
	for (arrp = arr; arrp < &arr[n - 1]; ++arrp) {
	    if (*arrp < pivot) {
		unsigned int tmpval = arr[pivindex];
		arr[pivindex++] = *arrp;
		*arrp = tmpval;
	    }
	}
	arr[n - 1] = arr[pivindex];
	arr[pivindex] = pivot;
	quicksort(arr, pivindex);
	++pivindex;
	quicksort(arr + pivindex, n - pivindex);
}


/*
 *	Probe to find Synaptics touchpad.
 */

#if HAVE_LINUX_INPUT_H

#define	INPUTNAME "SynPS/2 Synaptics TouchPad"

static const char	input_name[]	= INPUTNAME;

static Bool
syn_probe_path(const char *path)
{
	char	buf[64];
	int	fd;

	fd = open(path, O_RDONLY);
	if (fd == -1) {
	    perror(path);
	    return FALSE;
	}

	if (!ioctl(fd, EVIOCGNAME(sizeof buf), buf)) {
	    close(fd);
	    fputs("ioctl on ", stderr);
	    perror(path);
	    return FALSE;
	}

	close(fd);

	if (strlen(buf) < sizeof input_name - 1
	  || memcmp(input_name, buf, sizeof input_name) != 0)
	    return FALSE;	/* no match */

	printf("Found \"%s\" on %s\n", buf, path);

	return TRUE;
}

static Bool
syn_probe_num(InputInfoPtr infp, unsigned int n)
{
	char	path[32];	/* /dev/input/eventnnnnn. */

	if (snprintf(path, sizeof path, "/dev/input/event%u", n) >= sizeof path)
	    return FALSE;	/* internal error -- path too short */

	if (syn_probe_path(path)) {
	    xf86Msg(X_PROBED, "%s: found Synaptics touchpad on %s\n",
	      infp->name, path);
	    infp->options = xf86ReplaceStrOption(infp->options, "Device", path);
	    return TRUE;
	}

	return FALSE;
}

static Bool
syn_probe(InputInfoPtr infp)
{
	DIR		*dir;
	unsigned int	i;
	unsigned int	mask	= 0;
	unsigned int	*arr	= NULL;
	size_t		arrsize	= 0;
	size_t		maxsize	= 0;
	unsigned int	mask1;

	dir = opendir("/dev/input");
	if (dir == NULL) {
	    perror("/dev/input");
	    return FALSE;
	}

	for (;;) {
	    struct dirent *dp = readdir(dir);
	    const char *cp;

	    if (dp == NULL)
		break;

	    if (memcmp(dp->d_name, "event", 5) != 0)
		continue;

	    cp = dp->d_name + 5;
	    i = (unsigned char) (*cp - '0');
	    if (i >= 10) continue;

	    while (((unsigned char) (*++cp - '0')) < 10)
		i = i * 10 + (*cp - '0');

	    if (*cp != '\0')
		continue;

	    if (i < 8 * sizeof mask)
		mask |= 1 << i;
	    else {
		if (arrsize >= maxsize) {
		    maxsize += 64;
		    arr = (arr == NULL ? malloc(maxsize * sizeof (*arr))
		      : realloc(arr, maxsize * sizeof (*arr)));
		    if (arr == NULL) {
			fprintf(stderr, "Error allocating %ld integers\n",
			  maxsize);
			return FALSE;
		    }
		}
		arr[arrsize++] = i;
	    }
	}

	closedir(dir);

	i = 0;
	mask1 = 1;
	while (mask != 0) {
	    if (mask & mask1) {
		if (syn_probe_num(infp, i)) {
		    if (arr != NULL)
			free(arr);
		    return TRUE;
		}
		mask ^= mask1;
	    }
	    ++i;
	    mask1 <<= 1;
	}

	quicksort(arr, arrsize);

	for (i = 0; i < arrsize; ++i) {
	    if (syn_probe_num(infp, arr[i])) {
		free(arr);
		return TRUE;
	    }
	}

	if (arr != NULL)
	    free(arr);

	return FALSE;
}

#endif /* HAVE_LINUX_INPUT_H */


/*========================================================================*\
 *
 *	Procedures for setting up driver
 *
\*========================================================================*/

static void
set_geom(SynptGeomPtr geom, struct int_geom *igeom)
{
	igeom->l = SYN_XMIN + (SYN_XSIZE * (int) geom->l + 255/2) / 255;
	igeom->r = SYN_XMAX - (SYN_XSIZE * (int) geom->r + 255/2) / 255;
	igeom->t = SYN_YMAX - (SYN_YSIZE * (int) geom->t + 255/2) / 255;
	igeom->b = SYN_YMIN + (SYN_YSIZE * (int) geom->b + 255/2) / 255;
}

static void
set_options(SynapticsPrivatePtr synp)
{
	if (synp->opt.flags & SYN_FAST) {
	    synp->rate_factor = 1;
	    synp->speed_smooth = 1.0 / 6;
	    synp->dist2speed = .04;
	}
	else {
	    synp->rate_factor = 2;
	    synp->speed_smooth = 1.0 / 3;
	    synp->dist2speed = .02;
	}

	set_geom(&synp->opt.mo_g, &synp->mo_ig);
	set_geom(&synp->opt.scrl_g, &synp->scrl_ig);
	set_geom(&synp->opt.cor_g, &synp->cor_ig);
}

static void
init_geom_options(pointer options,
  SynptGeomPtr geom, const SynptGeomRec *deflt_geom,
  const char *left_name, const char *right_name,
  const char *top_name, const char *bottom_name)
{
	double	f;

	f = xf86SetRealOption(options, left_name, deflt_geom->l / 255.0);
	geom->l = (f < 0.0 ? 0 : f > 1.0 ? 255 : (int) (f * 255));

	f = xf86SetRealOption(options, right_name, deflt_geom->r / 255.0);
	geom->r = (f < 0.0 ? 0 : f > 1.0 ? 255 : (int) (f * 255));

	f = xf86SetRealOption(options, top_name, deflt_geom->t / 255.0);
	geom->t = (f < 0.0 ? 0 : f > 1.0 ? 255 : (int) (f * 255));

	f = xf86SetRealOption(options, bottom_name, deflt_geom->b / 255.0);
	geom->b = (f < 0.0 ? 0 : f > 1.0 ? 255 : (int) (f * 255));
}

static InputInfoPtr
pre_init(InputDriverPtr drv, IDevPtr dev, int flags)
{
	InputInfoPtr		infp;
	SynapticsPrivatePtr	synp;
	ExtensionEntry		*entry;
	const struct optl	*optp;
	char			*str;

	infp = xf86AllocateInput(drv, 0);
	if (infp == NULL)
	    return NULL;

	infp->name = dev->identifier;
	infp->flags = XI86_POINTER_CAPABLE | XI86_SEND_DRAG_EVENTS;

	/* Procedures */
	infp->device_control = tp_control;
	infp->read_input = tp_read;
	infp->control_proc = NULL;
	infp->close_proc = NULL;
	infp->switch_mode = NULL;
	infp->conversion_proc = tp_convert;
	infp->reverse_conversion_proc = NULL;

	/* Data */
	infp->fd = -1;
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
	infp->motion_history_proc = xf86GetMotionEvents;
#endif
	infp->type_name = XI_TOUCHPAD;
	infp->conf_idev = dev;

	/* Private area */
	synp = xcalloc(1, sizeof *synp);
	if (synp == NULL) {
	    xf86Msg(X_ERROR, "%s: cannot reserve memory for state record\n",
	      infp->name);
	    return infp;
	}
	infp->private = synp;
	synp->infp = infp;

	/* Options processing */
	xf86CollectInputOptions(infp, NULL, NULL);
	xf86ProcessCommonOptions(infp, infp->options);

	str = xf86SetStrOption(infp->options, "Protocol", NULL);
	if (str != NULL) {
	    if (xf86NameCmp(str, "auto-dev") == 0)
		synp->protocol = SYN_PROTO_AUTO;
	    else if (xf86NameCmp(str, "psaux") == 0)
		synp->protocol = SYN_PROTO_RAW;
#if HAVE_LINUX_INPUT_H
	    else if (xf86NameCmp(str, "event") == 0)
		synp->protocol = SYN_PROTO_EVENT;
#endif
	    else if (xf86NameCmp(str, "psm") == 0)
		synp->protocol = SYN_PROTO_PSM;
	    else {
		xf86Msg(X_ERROR, "%s: Invalid prototype name %s\n", infp->name,
		  str);
		xfree(str);
		xfree(synp);
		infp->private = NULL;
		return infp;
	    }
	    xfree(str);
	}

	for (optp = optlist; optp < optlist + Number(optlist); ++optp) {
	    switch (optp->type) {
		case OPT_FLAG:
		    if (xf86SetBoolOption(infp->options, optp->name,
		      optp->deflt))
			synp->opt.flags |= optp->value;
		    break;

		case OPT_BYTE:
		    {
			int		i;
			unsigned char	c;

			i = xf86SetIntOption(infp->options, optp->name,
			  optp->deflt);
			c = (i < optp->min ? optp->min
			  : (i > optp->max ? optp->max : i));
			memcpy((char *) &synp->opt + optp->value, &c, sizeof c);
		    }
		    break;

		case OPT_SHORT:
		    {
			int		i;
			unsigned short	s;

			i = xf86SetIntOption(infp->options, optp->name,
			  optp->deflt);
			s = (i < optp->min ? optp->min
			  : (i > optp->max ? optp->max : i));
			memcpy((char *) &synp->opt + optp->value, &s, sizeof s);
		    }
		    break;
	    }
	}

	init_geom_options(infp->options, &synp->opt.mo_g,
	  &default_motion_geometry, "MotionPosLeft", "MotionPosRight",
	  "MotionPosTop", "MotionPosBottom");

	init_geom_options(infp->options, &synp->opt.scrl_g,
	  &default_scroll_geometry, "ScrollPosLeft", "ScrollPosRight",
	  "ScrollPosTop", "ScrollPosBottom");

	init_geom_options(infp->options, &synp->opt.cor_g,
	  &default_corner_geometry, "CornerPosLeft", "CornerPosRight",
	  "CornerPosTop", "CornerPosBottom");

	switch (synp->protocol) {

	    case SYN_PROTO_AUTO:
		str = xf86SetStrOption(infp->options, "Device", NULL);
		if (str == NULL) {
#if HAVE_LINUX_INPUT_H
		    if (syn_probe(infp)) {
			synp->protocol = SYN_PROTO_EVENT;
			synp->opt.flags |= SYN_FAST;
		    }
		    else
#endif
		    {
			synp->protocol = SYN_PROTO_RAW;
			infp->options = xf86ReplaceStrOption(infp->options,
			  "Device", "/dev/psaux");
		    }
		}
		else {
#if HAVE_LINUX_INPUT_H
		    if (syn_probe_path(str)) {
			synp->protocol = SYN_PROTO_EVENT;
			synp->opt.flags |= SYN_FAST;
		    }
		    else
#endif
			synp->protocol = SYN_PROTO_RAW;

		    xfree(str);
		}
		break;

	    case SYN_PROTO_RAW:
		infp->options = xf86ReplaceStrOption(infp->options, "Device",
		  "/dev/psaux");
		break;

#if HAVE_LINUX_INPUT_H
	    case SYN_PROTO_EVENT:
		{
		    Bool ret;

		    str = xf86SetStrOption(infp->options, "Device", NULL);
		    if (str == NULL)
			ret = syn_probe(infp);
		    else {
			ret = syn_probe_path(str);
			xfree(str);
		    }
		    if (!ret) {
			xf86Msg(X_ERROR, "%s: Synaptics touchpad not found\n",
			  infp->name);
			xfree(synp);
			infp->private = NULL;
			return infp;
		    }
		    synp->opt.flags |= SYN_FAST;
		}
		break;
#endif

	    case SYN_PROTO_PSM:
		xf86Msg(X_ERROR, "%s: psm protocol is not yet supported\n",
		  infp->name);
		xfree(synp);
		infp->private = NULL;
		return infp;
	}

	synp->protobak = synp->protocol;

	set_options(synp);

	infp->fd = xf86OpenSerial(infp->options);
	if (infp->fd == -1) {
	    xf86Msg(X_ERROR, "%s: cannot open input device\n", infp->name);
	    xfree(synp);
	    infp->private = NULL;
	    return infp;
	}
	xf86CloseSerial(infp->fd);
	infp->fd = -1;

	infp->flags |= XI86_CONFIGURED;

	/*
	 * I don't use LoadExtension (hw/xfree86/loader/loadext.c) here
	 * because I don't see a way to make the extension and the input
	 * module aware of each other.  Besides, we're not really loading it,
	 * the input module is already loaded.
	 */
	entry = AddExtension(SynptExtensionName,
	  SynptNumberEvents, SynptNumberErrors,
	  extn_main, extn_swapped_main, extn_closedown, StandardMinorOpcode);

	if (entry == NULL)
	    xf86Msg(X_WARNING, "%s: cannot create server extension "
	      SynptExtensionName "\n", infp->name);
	else {
	    synp->extn = entry;
	    entry->extPrivate = synp;
	}

	return infp;
}

static pointer
module_setup(pointer module, pointer options, int *errmajor, int *errminor)
{
	xf86AddInputDriver(&synaptics, module, 0);
	return module;
}

static void
module_teardown(pointer p)
{
}

/*
 *	If I don't have this here, then "xset q" crashes the server, and I don't
 *	get the functionality of the mouse acceleration provided therein.
 */

static void
pointer_control(DeviceIntPtr device, PtrCtrl *ctrl)
{
}

static Bool
tp_control(DeviceIntPtr device, int what)
{
	InputInfoPtr		infp	= device->public.devicePrivate;
	SynapticsPrivatePtr	synp	= infp->private;
	unsigned char		map[SYN_N_BUTTONS + 1];
	int			i;

	switch (what) {

	    case DEVICE_INIT:
		/* Here we have information on the screen size, if we want it */

		if (device->public.on) {
		    ErrorF("Synaptics DEVICE_INIT: device was on??\n"); /*|||*/
		    device->public.on = FALSE;
		}

		for (i = 0; i <= SYN_N_BUTTONS; ++i)
		    map[i] = i;

		if (!InitPointerDeviceStruct((DevicePtr) device, map,
		  SYN_N_BUTTONS,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
		  miPointerGetMotionEvents,
#elif GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 3
		  GetMotionHistory,
#endif
		  pointer_control,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
		  miPointerGetMotionBufferSize()
#else
		  GetMotionHistorySize(), 2
#endif
		  )) {
		    ErrorF("Synaptics DEVICE_INIT: could not initialize pointer struct\n");
		    return !Success;
		}

		/* X valuator */
		xf86InitValuatorAxisStruct(device, 0, 0, -1, 1, 0, 1);
		xf86InitValuatorDefaults(device, 0);
		/* Y valuator */
		xf86InitValuatorAxisStruct(device, 1, 0, -1, 1, 0, 1);
		xf86InitValuatorDefaults(device, 1);

		/* Motion history */
		xf86MotionHistoryAllocate(infp);

		break;

	    case DEVICE_ON:
		infp->fd = xf86OpenSerial(infp->options);
		if (infp->fd == -1) {
		    xf86Msg(X_WARNING, "%s: cannot open input device\n",
		      infp->name);
		    return !Success;
		}

		synp->buffer = XisbNew(infp->fd, 64);
		if (synp->buffer == NULL) {
		    xf86CloseSerial(infp->fd);
		    infp->fd = -1;
		    return !Success;
		}

		switch (synp->protocol) {

		    case SYN_PROTO_RAW:
			if (!init_synaptics(infp)) {
			    xf86CloseSerial(infp->fd);
			    infp->fd = -1;
			    XisbFree(synp->buffer);
			    synp->buffer = NULL;
			    return !Success;
			}
			break;

#if HAVE_LINUX_INPUT_H
		    case SYN_PROTO_EVENT:
			/* Prevent touchpad events from being sent on to the
			   /dev/input/mice interface (and being handled again
			   by the mouse driver).  */
			if (ioctl(infp->fd, EVIOCGRAB, 1) == -1)
			    xf86Msg(X_WARNING,
			      "%s: cannot grab event interface (errno=%d)\n",
			      infp->name, errno);
			break;
#endif
		}

		xf86FlushInput(infp->fd);
		xf86AddEnabledDevice(infp);

		synp->in_sync = 1;

		device->public.on = TRUE;
		break;

	    case DEVICE_OFF:
	    case DEVICE_CLOSE:
		if (infp->fd != -1) {
		    if (synp->emu3timer != NULL) {
			TimerFree(synp->emu3timer);
			synp->emu3timer = NULL;
		    }
		    if (synp->gesture_timer != NULL) {
			TimerFree(synp->gesture_timer);
			synp->gesture_timer = NULL;
		    }
		    synp->emu3state = 0;
		    synp->gbits = synp->hard_buttons = synp->buttons = 0;

		    switch (synp->protocol) {
			case SYN_PROTO_RAW: {
			    static const unsigned char s1[] = {
				    246,	/* set defaults and disable */
				    230,	/* set slow rate */
				    244 };	/* enable */

			    ps2_send(infp, s1, sizeof s1,
			      "Synaptics touchpad flagged error on shutdown\n");
			    break;
			}

#if HAVE_LINUX_INPUT_H
			case SYN_PROTO_EVENT:
			    if (ioctl(infp->fd, EVIOCGRAB, 0) == -1)
				xf86Msg(X_WARNING,
				  "%s: cannot ungrab event interface (errno=%d)\n",
				  infp->name, errno);
			    break;
#endif
		    }

		    xf86RemoveEnabledDevice(infp);
		    if (synp->buffer) {
			XisbFree(synp->buffer);
			synp->buffer = NULL;
		    }
		    xf86CloseSerial(infp->fd);
		    infp->fd = -1;
		}
		device->public.on = FALSE;
		break;

	    default:
		ErrorF("Synaptics_drv: unknown code %d\n", what);
		return !Success;
	}

	return Success;
}


/*
 *	3-button emulation tables.  Since the touchpad gives us a steady
 *	stream of 40 mouse events per second whenever anything at all is going
 *	on, we count mouse events as a basis for doing timeouts.
 *	In all other respects, the algorithm here should give the same results
 *	as the finite state machine in mouse.c.
 *
 *	As in mouse.c, we do 3-button emulation via a finite-state machine.
 *	Here is our list of (11) states:
 *
 *	0  ground		initial state, showing no buttons
 *	1  pressed left		left button pressed and showing
 *	2  pressed right	right button pressed and showing
 *	3  pressed both		both buttons pressed and showing
 *
 *	4  delayed left		left pressed, waiting for right, timeout set
 *	5  delayed right	right pressed, waiting for left, timeout set
 *
 *	6  middle pressed	both buttons pressed, middle button showing
 *
 *	7  released left	middle button showing, right button still down
 *	8  released right	middle button showing, left button still down
 *	9  repressed left	both buttons pressed, middle and left showing
 *	10 repressed right	both buttons pressed, middle and right showing
 *
 *
 *	The data structures are:
 *		state_next	Next state to go to
 *		state_show	Buttons to show now (8 --> set timeout)
 *
 *	Sometimes it is necessary to impose an action right now (e.g., a quick
 *	press of button 1 needs to generate an event somehow).  This is
 *	reflected in some upper bits of state_next.
 */

static unsigned char state_next[13][4] = {
		{0, 4, 5, 6},			/* 0 ground */
		{0, 1, 5, 3},			/* 1 pressed left */
		{0, 4, 2, 3},			/* 2 pressed right */
		{0, 1, 2, 3},			/* 3 pressed both */
		{(1<<4) | 0, 4, (1<<4) | 5, 6},	/* 4 delayed left */
		{(3<<4) | 0, (3<<4) | 4, 5, 6},	/* 5 delayed right */
		{0, 8, 7, 6},			/* 6 middle pressed */
		{0, 4, 7, 9},			/* 7 released left */
		{0, 8, 5, 10},			/* 8 released right */
		{0, 1, 7, 9},			/* 9 repressed left */
		{0, 8, 2, 10},			/* 10 repressed right */
};

static unsigned char state_show[13] = {0, 1, 4, 5, 8, 8, 2, 2, 2, 3, 6};


CARD32
emu3timerfunc(OsTimerPtr timer, CARD32 now, pointer arg)
{
	SynapticsPrivatePtr	synp	= (SynapticsPrivatePtr) arg;
	int			button;

	synp->emu3state -= 3;	/* 4 --> 1, 5 --> 2 */

	button = state_show[synp->emu3state];

	if (!(synp->buttons & button)) {
	    int wasset = xf86BlockSIGIO();

	    xf86PostButtonEvent(synp->infp->dev, 0, button - button / 4,
	      TRUE, 0, 0);
	    synp->buttons |= button;
	    xf86UnblockSIGIO(wasset);
	}

	synp->hard_buttons |= button;
	synp->emu3timer = NULL;

	return 0;
}


CARD32
gesture_func(OsTimerPtr timer, CARD32 now, pointer arg)
{
	SynapticsPrivatePtr	synp	= (SynapticsPrivatePtr) arg;
	unsigned char		changes;

	synp->gesture_timer = NULL;

	if (synp->pressing && synp->opt.flags & SYN_TAPDRAG)
	    return 0;

	synp->gbits = 0;
	changes = synp->buttons ^ synp->hard_buttons;

	if (changes != 0) {
	    int button = 1;
	    unsigned char mask = 1;
	    int wasset = xf86BlockSIGIO();

	    for (;;) {
		if (changes & mask) {
		    xf86PostButtonEvent(synp->infp->dev, 0, button, FALSE,
		      0, 0);
		    changes ^= mask;
		    if (changes == 0)
			break;
		}
		++button;
		mask <<= 1;
	    }
	    xf86UnblockSIGIO(wasset);
	}

	synp->buttons = synp->hard_buttons;

	return 0;
}


/* Determine which button to use in a tap gesture.  */

static int
which_button(SynapticsPrivatePtr synp)
{
	int	btn;

	/* Multi finger taps take precedence (if enabled) */

	if (synp->max_fingers > 1) {
	    btn = synp->opt.tap_button[synp->max_fingers - 1];
	    if (btn != 0)
		return btn;
	}

	/* Next check for corner taps (if enabled) */

	if (synp->gy > synp->cor_ig.t) {
	    if (synp->gx < synp->cor_ig.l) {
		btn = synp->opt.ul_button;
		if (btn != 0)
		    return btn;
	    }
	    else if (synp->gx > synp->cor_ig.r) {
		btn = synp->opt.ur_button;
		if (btn != 0)
		    return btn;
	    }
	}
	else if (synp->gy < synp->cor_ig.b) {
	    if (synp->gx < synp->cor_ig.l) {
		btn = synp->opt.ll_button;
		if (btn != 0)
		    return btn;
	    }
	    else if (synp->gx > synp->cor_ig.r) {
		btn = synp->opt.lr_button;
		if (btn != 0)
		    return btn;
	    }
	}

	/* Else, return 0 if multi finger (already determined), or the
	   appropriate button for single finger.  */

	return synp->max_fingers > 1 ? 0 : synp->opt.tap_button[0];
}


static void
tp_read(InputInfoPtr infp)
{
	SynapticsPrivatePtr	synp	= infp->private;
	int			value;
	int			dx, dy;
	int			buttons;
	unsigned char		gbits;		/* gesture bits _this time_ */
	unsigned char		wheel_bits;
	int			scroll_dx, scroll_dy;
	int			change;
	int			i;

	/*
	 * We already know there's at least one byte of input, so set blocking
	 * to -1 for now.  This is automatically cleared after the first byte
	 * so that successive reads do not block (this is required).
	 */
	XisbBlockDuration(synp->buffer, -1);

	for (;;) {
	    value = XisbRead(synp->buffer);
	    if (value < 0)	/* if no more input */
		break;

	    switch (synp->protocol) {

		case SYN_PROTO_RAW:

		    synp->u.packet[synp->packet_size++] = (unsigned char) value;

		    /* ErrorF("Synaptics byte:  0x%02x\n", value); */

		    if (synp->packet_size < 6)
			continue;

		    /* Packet received.  Decode it.  */

		    /* ErrorF("Mouse packet:  %02x %02x %02x %02x %02x %02x\n",
		      synp->u.packet[0], synp->u.packet[1], synp->u.packet[2],
		      synp->u.packet[3], synp->u.packet[4], synp->u.packet[5]);
		    */
		    synp->packet_size = 0;

		    /* Check for invalid packet.  */
		    if ((synp->u.packet[0] & 0xc8) != 0x80
		      || (synp->u.packet[3] & 0xc8) != 0xc0) {	/* if invalid */
			if (synp->in_sync) {
			    synp->in_sync = 0;
			    ErrorF("Synaptics driver lost sync\n");
			}
#ifdef EXTMOUSEDEBUG
			ErrorF("Skipping byte %02x\n", synp->buffer[0]);
#endif
			for (i = 0; i < 5; ++i)
			    synp->u.packet[i] = synp->u.packet[i + 1];
			synp->packet_size = 5;

			continue;
		    }

		    if (!synp->in_sync) {
			synp->in_sync = 1;
			ErrorF("Synaptics driver back in sync\n");
		    }

		    synp->x = ((synp->u.packet[3] & 0x10) << 8)
		      | ((synp->u.packet[1] & 0x0f) << 8) | synp->u.packet[4];
		    synp->y = ((synp->u.packet[3] & 0x20) << 7)
		      | ((synp->u.packet[1] & 0xf0) << 4) | synp->u.packet[5];
		    synp->z = synp->u.packet[2];

		    synp->w = 5;	/* ||| w is not used yet */
		    if (synp->cap_bits)
			synp->w = ((synp->u.packet[0] & 0x30) >> 2)
			  | ((synp->u.packet[0] & 0x04) >> 1)
			  | ((synp->u.packet[3] & 0x04) >> 2);
		    synp->ev_fingers = (synp->w < 2 ? synp->w + 2 : 1);

		    synp->ev_buttons = ((synp->u.packet[0] ^ synp->u.packet[3])
		      & 0x3) << 2
		      | (synp->u.packet[0] & 0x03);

		    break;

#if HAVE_LINUX_INPUT_H
		case SYN_PROTO_EVENT:

		    i = 0;
		    do {
			synp->u.packet[i++] = (unsigned char) value;
			if (i >= sizeof (struct input_event))
			    break;
			value = XisbRead(synp->buffer);
		    } while (value >= 0);

		    if (i < sizeof (struct input_event)) {
			ErrorF(
			  "Synaptics event driver received incomplete packet\n"
			  );
			continue;
		    }

		    switch (synp->u.ev.type) {
			case EV_SYN:
			    if (synp->u.ev.code == SYN_REPORT)
				break;	/* packet is complete */
			    continue;

			case EV_KEY:
			    switch (synp->u.ev.code) {
				case BTN_LEFT:
				    if (synp->u.ev.value)
					synp->ev_buttons |= SYN_BTN_1;
				    else
					synp->ev_buttons &= ~SYN_BTN_1;
				    break;
				case BTN_RIGHT:
				    if (synp->u.ev.value)
					synp->ev_buttons |= SYN_BTN_2;
				    else
					synp->ev_buttons &= ~SYN_BTN_2;
				    break;
				case BTN_FORWARD:
				    if (synp->u.ev.value)
					synp->ev_buttons |= SYN_BTN_U;
				    else
					synp->ev_buttons &= ~SYN_BTN_U;
				    break;
				case BTN_BACK:
				    if (synp->u.ev.value)
					synp->ev_buttons |= SYN_BTN_D;
				    else
					synp->ev_buttons &= ~SYN_BTN_D;
				    break;
				case BTN_TOUCH:
				    break;
				case BTN_TOOL_FINGER:
				    synp->ev_fingers =
				      (synp->u.ev.value ? 1 : 0);
				    break;
				case BTN_TOOL_DOUBLETAP:
				    if (synp->u.ev.value) synp->ev_fingers = 2;
				    break;
				case BTN_TOOL_TRIPLETAP:
				    if (synp->u.ev.value) synp->ev_fingers = 3;
				    break;
				/* default:
				    printf("EV_KEY 0x%x 0x%x\n",
				      synp->u.ev.code, synp->u.ev.value); */
			    }
			    continue;

			case EV_ABS:
			    switch (synp->u.ev.code) {
				case ABS_X:
				    synp->x = synp->u.ev.value;
				    break;
				case ABS_Y:
				    synp->y = SYN_YMAX + SYN_YMIN
				      - synp->u.ev.value;
				    break;
				case ABS_PRESSURE:
				    synp->z = synp->u.ev.value;
				    break;
				case ABS_TOOL_WIDTH:
				    synp->w = synp->u.ev.value;
				    break;
				/* default:
				    printf("EV_ABS %u %u\n", synp->u.ev.code,
				      synp->u.ev.value); */
			    }
			    continue;

			default:
			    /* printf("Unrecognized type 0x%x\n", ep->type); */
			    continue;
		    }
		    break;
#endif

		case SYN_PROTO_OFF:
		    continue;

	    } /* end switch over protocols */

	    /* --- begin extractable part --- */

	    dx = dy = 0;	/* set default behavior */
	    gbits = synp->gbits;

	    scroll_dx = scroll_dy = 0;
	    if (synp->ev_buttons & ~synp->prev_ev_buttons & SYN_BTN_U)
		scroll_dy = -1;
	    if (synp->ev_buttons & ~synp->prev_ev_buttons & SYN_BTN_D)
		++scroll_dy;
	    synp->prev_ev_buttons = synp->ev_buttons;

	    /*
	     * gesture_timer, gbits, and pressing are as follows:
	     *
	     *  timer	 gbits	pressing
	     *	 0	   0	  any		no gesture in progress
	     *	 0	 != 0	   1		drag part of tap and drag
	     *	!= 0	   0	   1		finger down, possible tap
	     *	!= 0	 != 0	   0		sending tap, awaiting tap + drag
	     *	!= 0	 != 0	   1		possible tap and drag or 2nd tap
	     */

	    if (synp->pressing) {
		if (synp->z < synp->opt.finger_low) {
			/* if breaking finger contact */
		    synp->pressing = 0;
		    if (synp->gesture_timer != NULL
		      && (i = which_button(synp)) != 0) {
			/* if tap in progress */
			if (i >= 4 && i <= 7) {	/* if scrolling button */
			    /* Do 0-length taps for these buttons so they do
			       not interfere with scrolling by other means */
			    xf86PostButtonEvent(infp->dev, 0, i, TRUE, 0, 0);
			    xf86PostButtonEvent(infp->dev, 0, i, FALSE, 0, 0);
			    TimerFree(synp->gesture_timer);
			    synp->gesture_timer = NULL;
			}
			else {
			    synp->gbits = 1 << (i - 1);
			    if (gbits)		/* double click */
				gbits = 0;	/* unclick just this once */
			    else
				gbits = synp->gbits;

			    /* Set timer for ending the click */
			    synp->gesture_timer = TimerSet(synp->gesture_timer,
			      0, synp->opt.time_tapdur, gesture_func, synp);
			}
		    } else {
			gbits = synp->gbits = 0; /* if tap and drag, end it */
		    }
		} else {	/* if finger stays down */
		    double st;	/* absolute units to mickey conversion factor */

		    dx = synp->x - synp->prev_x;
		    dy = synp->prev_y - synp->y;

		    if (synp->ev_fingers != synp->fingers) {
			synp->fingers = synp->ev_fingers;
			if (synp->fingers > synp->max_fingers)
			    synp->max_fingers = synp->fingers;
			dx = dy = 0;
			if (synp->fingers == 2) {
			    if (synp->opt.flags
			      & (SYN_H_2FNG_SCRL | SYN_V_2FNG_SCRL)) {
				/* begin 2-finger scrolling */
				synp->scroll_type =
				  ((synp->opt.flags & SYN_H_2FNG_SCRL)
				  ? SCROLL_H_2FNGRS : 0)
				  | ((synp->opt.flags & SYN_V_2FNG_SCRL)
				  ? SCROLL_V_2FNGRS : 0);
				synp->dx_low_bits = synp->dy_low_bits = 1 << 7;
			    }
			}
			else
			    if (synp->scroll_type
			      & (SCROLL_H_2FNGRS | SCROLL_V_2FNGRS)) {
				/* end 2-finger scrolling */
				synp->scroll_type = 0;
				synp->dx_low_bits = synp->dy_low_bits = 1 << 7;
			    }
		    }

		    /* Turn off edge scrolling when the finger leaves the
		       edge. */

		    if (((synp->scroll_type & SCROLL_V_LEFT)
		      && synp->x >= synp->scrl_ig.l)
		      || ((synp->scroll_type & SCROLL_V_RIGHT)
		      && synp->x <= synp->scrl_ig.r)) {
			synp->scroll_type &= ~(SCROLL_V_LEFT | SCROLL_V_RIGHT);
			synp->dx_low_bits = synp->dy_low_bits = 1 << 7;
		    }

		    if (((synp->scroll_type & SCROLL_H_TOP)
		      && synp->y <= synp->scrl_ig.t)
		      || ((synp->scroll_type & SCROLL_H_BOTTOM)
		      && synp->y >= synp->scrl_ig.b)) {
			synp->scroll_type &= ~(SCROLL_H_TOP | SCROLL_H_BOTTOM);
			synp->dx_low_bits = synp->dy_low_bits = 1 << 7;
		    }

		    /*
		     * Edge motion.
		     */

		    /* This expression also tapers the effect as you get away
		       from the edge. */
#define DIST(V1, V2) \
	((int) ((synp->z <= synp->opt.edge_min_z \
	  ? (int) synp->opt.edge_min_speed \
	  : synp->z >= synp->opt.edge_max_z \
	    ? (int) synp->opt.edge_max_speed \
	    : (int) synp->opt.edge_min_speed \
	      + (double) (synp->opt.edge_max_speed - synp->opt.edge_min_speed) \
		  * (synp->z - synp->opt.edge_min_z) \
		  / (synp->opt.edge_max_z - synp->opt.edge_min_z)) \
	  * synp->rate_factor \
	  * ((3 * (V1) >= 2 * (V2)) ? 1.0 : 1.5 * (V1) / (V2)) + 0.5))

		    if (gbits || (synp->opt.flags & SYN_EDGEMOALWYS)) {
			if (!(synp->scroll_type & SCROLL_V_ALL)) {
			    if (synp->x < synp->mo_ig.l)
				dx = -DIST(synp->mo_ig.l - synp->x,
				  synp->mo_ig.l - SYN_XMIN);
			    else if (synp->x > synp->mo_ig.r)
				dx = DIST(synp->x - synp->mo_ig.r,
				  SYN_XMAX - synp->mo_ig.r);
			}
			if (!(synp->scroll_type & SCROLL_H_ALL)) {
			    if (synp->y < synp->mo_ig.b)
				dy = DIST(synp->mo_ig.b - synp->y,
				  synp->mo_ig.b - SYN_YMIN);
			    else if (synp->y > synp->mo_ig.t)
				dy = -DIST(synp->y - synp->mo_ig.t,
				  SYN_YMAX - synp->mo_ig.t);
			}
		    }
#undef DIST

		    /* st and synp->speed are independent of data rate */
		    st = synp->dist2speed * sqrt((double) (dx * dx + dy * dy));
		    if (st > 2) st = 2;
		    synp->speed += (st - synp->speed) * synp->speed_smooth;
		    st = synp->speed * (256 * .13/2);
		    if (st < 256 * .04) st = 256 * .04;

		    if (!(synp->scroll_type)) {
			synp->dx_low_bits = dx = (int) (dx * st + 0.5)
			  + synp->dx_low_bits;
			synp->dy_low_bits = dy = (int) (dy * st + 0.5)
			  + synp->dy_low_bits;
			dx >>= 8;
			dy >>= 8;
		    }
		    else {
			if ((synp->scroll_type & SCROLL_V_ALL)
			  && !(synp->scroll_type & SCROLL_H_EDGE)) {
			    synp->dy_low_bits = scroll_dy =
			      (int) (dy * st / synp->opt.v_scroll_delta + 0.5)
			      + synp->dy_low_bits;
			    scroll_dy >>= 8;
			}
			if ((synp->scroll_type & SCROLL_H_ALL)
			  && !(synp->scroll_type & SCROLL_V_EDGE)) {
			    synp->dx_low_bits = scroll_dx =
			      (int) (dx * st / synp->opt.h_scroll_delta + 0.5)
			      + synp->dx_low_bits;
			    scroll_dx >>= 8;
			}

			dx = dy = 0;
		    }

		    synp->prev_x = synp->x;
		    synp->prev_y = synp->y;
		}
	    }
	    else {	/* if not pressing */
		if (synp->z >= synp->opt.finger_high) {
			/* if making finger contact */
		    synp->pressing = 1;
		    synp->fingers = synp->max_fingers = synp->ev_fingers;
		    synp->prev_x = synp->x;
		    synp->prev_y = synp->y;
		    synp->dx_low_bits = synp->dy_low_bits = 1 << 7;
		    synp->speed = 0;

		    /*
		     * Check for two-finger scrolling and edge scrolling.
		     */
		    if (synp->max_fingers == 2 && (synp->opt.flags
		      & (SYN_H_2FNG_SCRL | SYN_V_2FNG_SCRL))) {
			synp->scroll_type = ((synp->opt.flags & SYN_H_2FNG_SCRL)
			  ? SCROLL_H_2FNGRS : 0)
			  | ((synp->opt.flags & SYN_V_2FNG_SCRL)
			  ? SCROLL_V_2FNGRS : 0);
		    } else {	/* check for edge scrolling */
			synp->scroll_type = 0;
			if ((synp->opt.flags & SYN_L_EDGE_SCRL)
			  && synp->x < synp->scrl_ig.l)
			    synp->scroll_type = SCROLL_V_LEFT;
			else if ((synp->opt.flags & SYN_R_EDGE_SCRL)
			  && synp->x > synp->scrl_ig.r)
			    synp->scroll_type = SCROLL_V_RIGHT;

			/* If we start in a corner, then we could be scrolling
			   in either direction (or tapping). */
			if ((synp->opt.flags & SYN_T_EDGE_SCRL)
			  && synp->y > synp->scrl_ig.t)
			    synp->scroll_type |= SCROLL_H_TOP;
			else if ((synp->opt.flags & SYN_B_EDGE_SCRL)
			  && synp->y < synp->scrl_ig.b)
			    synp->scroll_type |= SCROLL_H_BOTTOM;
		    }

		    /*
		     * If gestures are enabled, set the timeout to check for
		     * (first or second) tap.
		     */

		    if (synp->opt.flags & (SYN_TAPCLICK | SYN_TAPDRAG)) {
			synp->gesture_timer = TimerSet(synp->gesture_timer, 0,
			  synp->opt.time_maxtap, gesture_func, synp);
			/* save position of tap (it may go away when the finger
			   is lifted) */
			synp->gx = synp->x;
			synp->gy = synp->y;
		    }
		}
	    }

	    buttons = synp->ev_buttons & 0x03;

	    if (synp->opt.flags & SYN_EMU3) {
		int	state = synp->emu3state;

		state = state_next[state][buttons];

		if (state >> 4 != 0) {
		    xf86PostButtonEvent(infp->dev, 0, state >> 4, TRUE, 0, 0);
		    synp->buttons |= 1 << ((state >> 4) - 1);
		    state &= 0x0f;
		}

		buttons = state_show[state];

		if (buttons == 8) {
		    buttons = 0;
		    if (synp->emu3timer == NULL || state != synp->emu3state)
			synp->emu3timer = TimerSet(synp->emu3timer, 0,
			  synp->opt.time_emu3, emu3timerfunc, synp);
		}
		else {
		    TimerFree(synp->emu3timer);
		    synp->emu3timer = NULL;
		}

		synp->emu3state = state;
	    }
	    else
		buttons += buttons & 0x02;	/* shift bit 1 to bit 2 */

	    /* --- end extractable part --- */

	    synp->hard_buttons = buttons;

	    buttons |= gbits;			/* fold in gesture */

#ifdef SYNAPTICS_DEBUG
	    ErrorF("dx=%i dy=%i buttons=%x scroll_dy=%d\n", dx, dy, buttons,
	      scroll_dy);
#endif

	    if (dx || dy)
		xf86PostMotionEvent(infp->dev, 0, 0, 2, dx, dy);

	    change = buttons ^ synp->buttons;
	    if (change & 1)
		xf86PostButtonEvent(infp->dev, 0, 1, buttons & 1, 0, 0);
	    if (change & 2)
		xf86PostButtonEvent(infp->dev, 0, 2, buttons & 2, 0, 0);
	    if (change & 4)
		xf86PostButtonEvent(infp->dev, 0, 3, buttons & 4, 0, 0);
	    synp->buttons = buttons;

	    /* For wheel emulation, press and release.  */
	    if (scroll_dy < 0)
		do {
		    xf86PostButtonEvent(infp->dev, 0, 4, TRUE, 0, 0);
		    xf86PostButtonEvent(infp->dev, 0, 4, FALSE, 0, 0);
		} while (++scroll_dy < 0);
	    else
		while (scroll_dy-- > 0) {
		    xf86PostButtonEvent(infp->dev, 0, 5, TRUE, 0, 0);
		    xf86PostButtonEvent(infp->dev, 0, 5, FALSE, 0, 0);
		}

	    if (scroll_dx < 0)
		do {
		    xf86PostButtonEvent(infp->dev, 0, 6, TRUE, 0, 0);
		    xf86PostButtonEvent(infp->dev, 0, 6, FALSE, 0, 0);
		} while (++scroll_dx < 0);
	    else
		while (scroll_dx-- > 0) {
		    xf86PostButtonEvent(infp->dev, 0, 7, TRUE, 0, 0);
		    xf86PostButtonEvent(infp->dev, 0, 7, FALSE, 0, 0);
		}
	}
}

static Bool
tp_convert(InputInfoPtr infp,
	int first, int num, int v0, int v1, int v2, int v3, int v4, int v5,
	int *x, int *y)
{
	if (first == 0 && num == 2) {
	    *x = v0;
	    *y = v1;
	    return TRUE;
	}

	ErrorF("Synaptics tp_convert: first = %d, num = %d\n", first, num);
	return FALSE;
}

/*========================================================================*\
 *
 *	Stuff to set up an extension
 *
\*========================================================================*/

/*
 * Unfortunately, this routine seems to be needed, since the idea of installing
 * an extension from a loaded module was not anticipated in the server design.
 */

static SynapticsPrivatePtr
find_synp(void)
{
	ExtensionEntry *extn;

	extn = CheckExtension(SynptExtensionName);

	if (extn == NULL)
	    return NULL;
	else
	    return extn->extPrivate;
}

static void
swapo(SynptControlPtr optp)
{
	register int n;

	swapl(&optp->flags, n);
	swaps(&optp->time_emu3, n);
	swaps(&optp->time_maxtap, n);
	swaps(&optp->time_tapdur, n);
	swaps(&optp->h_scroll_delta, n);
	swaps(&optp->v_scroll_delta, n);
}

static int
ProcSynptGetVersion(ClientPtr client)
{
	/* REQUEST(xSynptGetVersionReq); */
	xSynptGetVersionReply rep;
	SynapticsPrivatePtr synp;
	register int n;

	REQUEST_SIZE_MATCH(xSynptGetVersionReq);

	synp = find_synp();
	if (synp == NULL)
	    return BadImplementation;

	rep.type = X_Reply;
	rep.length = 0;
	rep.sequenceNumber = client->sequence;
	rep.version = SynptVersion;
	rep.hw_version = synp->hw_version;
	rep.hw_subvers = synp->hw_subvers;
	rep.cap_bits = synp->cap_bits;
	rep.protocol = synp->protobak;
	if (client->swapped) {
	    swaps(&rep.sequenceNumber, n);
	    swapl(&rep.version, n);
	    swaps(&rep.cap_bits, n);
	}
	WriteToClient(client, sizeof(xSynptGetVersionReply), (char *) &rep);
	return client->noClientException;
}

static int
ProcSynptGetOptions(ClientPtr client)
{
	/* REQUEST(xSynptGetOptionsReq); */
	xSynptGetOptionsReply rep;
	SynapticsPrivatePtr synp;
	register int n;

	REQUEST_SIZE_MATCH(xSynptGetOptionsReq);

	synp = find_synp();
	if (synp == NULL)
	    return BadImplementation;

	rep.type = X_Reply;
	rep.length = (sizeof (xSynptGetOptionsReply) - sizeof (xGenericReply))
	  >> 2;
	rep.sequenceNumber = client->sequence;
	rep.opt = synp->opt;
	if (client->swapped) {
	    swaps(&rep.sequenceNumber, n);
	    swapo(&rep.opt);
	}
	WriteToClient(client, sizeof(xSynptGetOptionsReply), (char *) &rep);
	return client->noClientException;
}

static Bool
opt_valid(int protocol, const SynptControlPtr opt)
{
	if (opt->time_emu3 < 25 || opt->time_emu3 > 5000
	  || opt->time_maxtap < 25 || opt->time_maxtap > 5000
	  || opt->time_tapdur < 25 || opt->time_tapdur > 950
	  || opt->h_scroll_delta < 4 || opt->h_scroll_delta > 500
	  || opt->v_scroll_delta < 4 || opt->v_scroll_delta > 500
	  || opt->finger_low < 1 || opt->finger_high < 1
#if HAVE_LINUX_INPUT_H
	  || (protocol == SYN_PROTO_EVENT && !(opt->flags & SYN_FAST))
#endif
	  || opt->ul_button > SYN_N_BUTTONS || opt->ur_button > SYN_N_BUTTONS
	  || opt->ll_button > SYN_N_BUTTONS || opt->lr_button > SYN_N_BUTTONS
	  || opt->tap_button[0] > SYN_N_BUTTONS
	  || opt->tap_button[1] > SYN_N_BUTTONS
	  || opt->tap_button[2] > SYN_N_BUTTONS
	  || opt->ul_button > SYN_N_BUTTONS || opt->ur_button > SYN_N_BUTTONS
	  || opt->ll_button > SYN_N_BUTTONS || opt->lr_button > SYN_N_BUTTONS
	  )
	    return FALSE;

	return TRUE;
}

static void
tp_reinit(InputInfoPtr infp)
{
	SynapticsPrivatePtr	synp	= infp->private;
	int			value;

	/* reset read_input routine */

	infp->read_input = tp_read;
	/*
	 * We already know there's at least one byte of input, so set blocking
	 * to -1 for now.  This is automatically cleared after the first byte
	 * so that successive reads do not block.  (Everybody does this.)
	 */
	XisbBlockDuration(synp->buffer, -1);

	value = XisbRead(synp->buffer);

	if (value == 0xfe) {	/* resend */
	    static const unsigned char disable = 245;

	    if (ps2_send(infp, &disable, 1,
	      "Synaptics touchpad error on reinit"))
		value = 0xfa;	/* ack */
	}

	if (value == 0xfa)	/* if ack (and not value < 0, no input) */
	    if (!send_mode_enable(infp, "Synaptics touchpad error on reinit"))
		value = 0xfe;	/* flag error */

	if (value != 0xfa) {	/* if some sort of error */
	    synp->opt.flags ^= SYN_FAST;
	    if (value != 0xfe)
		ErrorF("Synaptics tp_reinit failure\n");
	}
}

static int
ProcSynptSetOptions(ClientPtr client)
{
	REQUEST(xSynptSetOptionsReq);
	SynapticsPrivatePtr synp;
	CARD32 changed_flags;

	REQUEST_SIZE_MATCH(xSynptSetOptionsReq);

	synp = find_synp();
	if (synp == NULL)
	    return BadImplementation;

	if (client->swapped)
	    swapo(&stuff->opt);

	if (!opt_valid(synp->protocol, &stuff->opt))	/* check validity */
	    return BadValue;

	changed_flags = synp->opt.flags ^ stuff->opt.flags;
	synp->opt = stuff->opt;

	/* Implement the changes */

	set_options(synp);

	if (changed_flags & SYN_ENABLED) {
	    if (synp->opt.flags & SYN_ENABLED)
		synp->protocol = synp->protobak;
	    else
		synp->protocol = SYN_PROTO_OFF;
	}

	if (changed_flags & SYN_FAST) {		/* SYN_PROTO_RAW only */
	    /* For some reason, ps2_send doesn't work from within extension
	     * routines, so we send a byte and expect an ack to trigger the
	     * rest of the reinitialization sequence.  */
	    static const unsigned char disable = 245;

	    synp->infp->read_input = tp_reinit;
	    xf86WriteSerial(synp->infp->fd, &disable, 1);
	}

	return client->noClientException;
}

static int
extn_main(ClientPtr client)
{
	REQUEST(xReq);		/* defined in include/dix.h */

	switch (stuff->data) {
	    case X_SynptGetVersion:
		return ProcSynptGetVersion(client);
	    case X_SynptGetOptions:
		return ProcSynptGetOptions(client);
	    case X_SynptSetOptions:
		return ProcSynptSetOptions(client);
	    default:
		return BadRequest;
	}
}

static int
SProcSynptGetVersion(ClientPtr client)
{
	register int n;
	REQUEST(xSynptGetVersionReq);

	swaps(&stuff->length, n);
	REQUEST_SIZE_MATCH(xSynptGetVersionReq);
	return ProcSynptGetVersion(client);
}

static int
SProcSynptGetOptions(ClientPtr client)
{
	register int n;
	REQUEST(xSynptGetOptionsReq);

	swaps(&stuff->length, n);
	REQUEST_SIZE_MATCH(xSynptGetOptionsReq);
	return ProcSynptGetOptions(client);
}

static int
SProcSynptSetOptions(ClientPtr client)
{
	register int n;
	REQUEST(xSynptSetOptionsReq);

	swaps(&stuff->length, n);
	REQUEST_SIZE_MATCH(xSynptSetOptionsReq);
	return ProcSynptSetOptions(client);
}

static int
extn_swapped_main(ClientPtr client)
{
	REQUEST(xReq);		/* defined in include/dix.h */

	switch (stuff->data) {
	    case X_SynptGetVersion:
		return SProcSynptGetVersion(client);
	    case X_SynptGetOptions:
		return SProcSynptGetOptions(client);
	    case X_SynptSetOptions:
		return SProcSynptSetOptions(client);
	    default:
		return BadRequest;
	}
}

static void
extn_closedown(ExtensionEntry *e)
{
}
