/*
 * Copyright (c) 2007  Paul Vojta
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 */

#define	X_SynptGetVersion	0
#define	X_SynptGetOptions	1
#define	X_SynptSetOptions	2

#define SynptNumberEvents	0
#define SynptNumberErrors	0

#define	SynptExtensionName	"Synaptics-Touchpad"
#define	SynptVersion		1234


/* Protocol symbols */

#define	SYN_PROTO_AUTO	0	/* should not occur */
#define	SYN_PROTO_RAW	1
#define SYN_PROTO_EVENT	2
#define	SYN_PROTO_PSM	3


typedef struct {
	CARD8		reqType;
	CARD8		synptReqType;		/* always X_SynptGetVersion */
	CARD16		length B16;
} xSynptGetVersionReq;
#define sz_xSynptGetVersionReq 4

typedef struct {
	BYTE		type;			/* X_Reply */
	CARD8		pad0;
	CARD16		sequenceNumber B16;
	CARD32		length B32;
	CARD32		version B32;
	CARD8		hw_version;
	CARD8		hw_subvers;
	CARD16		cap_bits B16;
	CARD8		protocol;
	CARD8		pad1a, pad1b, pad1c;
	CARD32		pad2 B32;
	CARD32		pad3 B32;
	CARD32		pad4 B32;
} xSynptGetVersionReply;
#define sz_xSynptGetVersionReply 32


/*
 *	Control record for communicating options
 */

typedef struct {
	CARD8		l, r, t, b;
} SynptGeomRec, *SynptGeomPtr;

/* When adding to this:  Always check that it's a multiple of 4 bytes,
 * and update all sz_* accordingly.  */

typedef struct {		/* Record for communication with front-end */
	CARD32		flags B32;
	CARD16		time_emu3 B16;
	CARD16		time_maxtap B16;
	CARD16		time_tapdur B16;
	CARD16		h_scroll_delta B16;
	CARD16		v_scroll_delta B16;
	CARD8		finger_high;
	CARD8		finger_low;
	CARD8		edge_min_z;
	CARD8		edge_max_z;
	CARD8		edge_min_speed;
	CARD8		edge_max_speed;
	CARD8		ul_button;
	CARD8		ur_button;
	CARD8		ll_button;
	CARD8		lr_button;
	CARD8		tap_button[3];
	CARD8		pad;
	SynptGeomRec	mo_g;
	SynptGeomRec	scrl_g;
	SynptGeomRec	cor_g;
} SynptControlRec, *SynptControlPtr;

/* Flag values for synp->opt.flags */
#define	SYN_ENABLED	(1<<0)		/* Enabled */
#define	SYN_FAST	(1<<1)		/* HighRate */
#define	SYN_EMU3	(1<<2)		/* Emulate3Buttons */
#define	SYN_TAPCLICK	(1<<3)		/* TapToClick */
#define	SYN_TAPDRAG	(1<<4)		/* TapAndDrag */
#define	SYN_EDGEMOALWYS	(1<<5)		/* EdgeMotionUseAlways */
#define	SYN_L_EDGE_SCRL	(1<<6)		/* LeftEdgeScroll */
#define	SYN_R_EDGE_SCRL	(1<<7)		/* RightEdgeScroll */
#define	SYN_T_EDGE_SCRL	(1<<8)		/* TopEdgeScroll */
#define	SYN_B_EDGE_SCRL	(1<<9)		/* BottomEdgeScroll */
#define	SYN_H_2FNG_SCRL	(1<<10)		/* HorizTwoFingerScroll */
#define	SYN_V_2FNG_SCRL	(1<<11)		/* VertTwoFingerScroll */


/*
 *	Request options
 */

typedef struct {
	CARD8		reqType;
	CARD8		synptReqType;		/* always X_SynptGetOptions */
	CARD16		length B16;
} xSynptGetOptionsReq;
#define sz_xSynptGetOptionsReq 4

typedef struct {
	BYTE		type;			/* X_Reply */
	CARD8		pad0;
	CARD16		sequenceNumber B16;
	CARD32		length B32;
	SynptControlRec	opt;			/* 10 words (currently) */
} xSynptGetOptionsReply;
#define sz_xSynptGetOptionsReply 48


/*
 *	Set options
 */

typedef struct {
	CARD8		reqType;
	CARD8		synptReqType;		/* always X_SynptSetOptions */
	CARD16		length B16;
	SynptControlRec	opt;			/* 10 words (currently) */
} xSynptSetOptionsReq;
#define sz_xSynptSetOptionsReq 44
